import html
import io
from typing import Optional, Set, Dict, List
from dataclasses import dataclass, field

import markupsafe
from tqdm import tqdm
from dataclasses_json import dataclass_json
import json
import os
import pathlib
import shlex
from loguru import logger
from enum import Enum

from flask import Blueprint, render_template, request, redirect, url_for, current_app, send_file
from werkzeug.utils import secure_filename

import config
from helpers.md_and_html_convertor import MdAndHTMLConvertor
import helpers.table_of_services
from helpers.file_utils import relative_path, extract_fid_from_filename, json_safe_load, generate_random_folder_name, \
    persistent_audit_scopes_file_path
from template_manager import PwndocTemplateManager
from helpers.flask_multifile_upload import save_multiple_uploaded_files
import template_pwndoc
from scan2report import scan2report
from api_templates import download_templates_from_pwndoc_to_scan2report
from pwndoc_api import login, get_findings_from_audit, upsert_raw_finding, get_audits
from template_scan2report import TemplateScan2Report
from template_pwndoc import get_fid_from_raw_finding
from template_converters import Scan2ReportConvert, HTML_ATTRIBUTES
from helpers.custom_logging import FlashLog


INFO_FILE = "_pwndoc_importer_processing.json"


class OutputFormat(Enum):
    PWNDOC = "pwndoc"
    JSON = "json"
    CSV = "csv"
    DOCX = "docx"


SEVERITIES = {  # todo: synchronize this with scan2report
    '0': 'Informative',
    '1': 'Low',
    '2': 'Medium',
    '3': 'High',
    '4': 'Critical'
}


# TODO: There is race condition. When two people are adding findings - the scan2report template synchronization might break the other scan2report run.

bp = Blueprint('api_process_findings', __name__, template_folder='templates')


def _get_audits() -> Dict[str, str]:
    audits_dict = {
        "Export to JSON": OutputFormat.JSON.value,
        "Export to CSV": OutputFormat.CSV.value,
        "Export to DOCX": OutputFormat.DOCX.value,
    }
    for audit_id, audit_name in get_audits():
        nice_name = f'{audit_name}  (id: {audit_id})'
        audits_dict[nice_name] = audit_id

    return audits_dict


def _list_profiles() -> List[str]:
    with open(relative_path("scan2report/plugins/profiles.json"), encoding="utf8") as f:
        data: dict = json.load(f)
    return list(data.keys())


@bp.route("/upload_scanner_result/<audit_id>", methods=["GET"])
@bp.route("/upload_scanner_result", methods=["GET"])
def upload_scanner_result_get(audit_id: str = ""):  # todo: maybe add additional arguments for setting persistence across error
    audits_dict = _get_audits()
    if len(audit_id) and audit_id not in list(audits_dict.values()):
        FlashLog.warning("Audit ID from URL doesn't exist.")
        audit_id = OutputFormat.DOCX.value

    languages = list(filter(lambda x: x != "og", config.LOCALES))
    profiles = _list_profiles()

    return render_template('upload_scanner_result_get.html', audit_dict=audits_dict, audit_id=audit_id, lang_list=languages, severities=SEVERITIES, profiles=profiles, allowed_scanner_extensions=config.ALLOWED_SCANNER_EXTENSIONS)


def _flash_msg_and_go_to_start(msg: str, level: str = "warning", audit_id: str = ""):
    FlashLog.log_with_level(msg, level, depth=2)
    return redirect(url_for('api_process_findings.upload_scanner_result_get', audit_id=audit_id))


@dataclass_json
@dataclass
class ProcessingSettings:
    audit_id: str = OutputFormat.DOCX.value
    upload_to_pwndoc: bool = False
    scan2report_args: str = ''
    __folder_name: str = field(default_factory=lambda: generate_random_folder_name("findings_import"))
    stored_files: Set[str] = field(default_factory=set)
    locale: str = 'cs'

    missing_templates: List[str] = field(default_factory=list)  # todo: missing templates might not include grouped findings which don't have their own fid in the resulting findings.json
    output_unspecified: dict = field(default_factory=dict)

    @staticmethod
    def __get_folder_path(folder_name: str) -> str:
        assert folder_name == secure_filename(folder_name)
        return os.path.join(current_app.config['UPLOAD_FOLDER'], folder_name)

    def get_folder_path(self) -> str:
        return self.__get_folder_path(self.__folder_name)

    def get_folder_name(self) -> str:
        return self.__folder_name

    @classmethod
    def __get_info_file_path(cls, folder_name: str) -> str:
        return os.path.join(cls.__get_folder_path(folder_name), INFO_FILE)

    @classmethod
    def load_from_folder_name(cls, folder_name: str) -> Optional['ProcessingSettings']:
        with open(cls.__get_info_file_path(folder_name), encoding="utf-8") as f:
            file_data = json.load(f)
            return ProcessingSettings.from_dict(file_data)

    def save(self):
        pathlib.Path(self.get_folder_path()).mkdir(exist_ok=True)
        with open(self.__get_info_file_path(self.__folder_name), "w", encoding="utf-8") as f:
            json.dump(self.to_dict(), f, indent=4)

    def cleanup_uploaded_files(self):
        files = [os.path.join(self.get_folder_path(), f) for f in self.stored_files]

        for file in files:
            if pathlib.Path(file).is_file():
                os.remove(file)

    def get_result_path(self) -> str:
        if self.audit_id == OutputFormat.DOCX.value:
            return self._get_result_docx_path()
        if self.audit_id == OutputFormat.CSV.value:
            return self._get_result_csv_path()
        return self._get_result_json_path()

    def _get_result_docx_path(self):
        return os.path.join(self.get_folder_path(), 'scan2report_output.docx')

    def _get_result_json_path(self):
        return os.path.join(self.get_folder_path(), 'scan2report_output.json')

    def _get_result_csv_path(self):
        return os.path.join(self.get_folder_path(), 'scan2report_output.csv')

    def _meta_file_path(self) -> str:
        return f"{self.get_result_path()}.meta.json"

    def load_meta_file(self) -> dict:
        meta_data = json_safe_load(self._meta_file_path())
        return meta_data if meta_data else {}


@bp.route("/upload_scanner_result", methods=["POST"])
def upload_scanner_result_post():
    audit_id = request.form.get("audit")
    locale = request.form.get("locale")
    profile = request.form.get("profile")
    min_severity = request.form.get("min-severity")

    def _flash_error_msg(msg: str, level: str = "warning"):
        return _flash_msg_and_go_to_start(msg, level, audit_id)  # note: audit_id is taken from outer scope

    if len(audit_id) and audit_id not in list(_get_audits().values()):
        return _flash_error_msg("This audit ID doesn't exist.")

    if locale not in config.LOCALES:
        return _flash_error_msg("This locale does not exist.")

    if profile not in _list_profiles():
        return _flash_error_msg("This profile does not exist.")

    if min_severity not in SEVERITIES.keys():
        return _flash_error_msg("This severity doesn't exist")

    scan2report_args = request.form.get("scan2report_args")
    scan2report_args += f' --min-severity {min_severity} --profile {profile}'  # todo: make sure that I'm not overwriting explicitly set parameters

    audit_info = list(filter(lambda x: x[0] == audit_id, get_audits()))
    upload_to_pwndoc = len(audit_info) > 0

    ps = ProcessingSettings(
        audit_id=audit_id,
        scan2report_args=scan2report_args,
        locale=locale,
        upload_to_pwndoc=upload_to_pwndoc,
    )

    error_msg, saved_filenames = save_multiple_uploaded_files(output_folder=ps.get_folder_path(), reserved_filenames=set(INFO_FILE), allowed_extensions=config.ALLOWED_SCANNER_EXTENSIONS)
    ps.stored_files = ps.stored_files.union(saved_filenames)
    ps.save()

    if len(error_msg) > 0:
        ps.cleanup_uploaded_files()
        return _flash_msg_and_go_to_start(error_msg, "error", audit_id)

    return redirect(url_for('api_process_findings.process_scanner_result_using_scan2report', folder_name=ps.get_folder_name()))


def _upload_missing_templates(ps: ProcessingSettings):
    PwndocTemplateManager.update_template_db_from_pwndoc()  # Update DB so that we can do automatic template merging.

    for single_missing_string in tqdm(ps.missing_templates):
        filename, severity, title = single_missing_string.strip().split("\t")
        filename = secure_filename(filename)
        fid = extract_fid_from_filename(filename)
        path = os.path.join(TemplateScan2Report.native_folder_name_for_locale(ps.locale, original=True), filename)
        with open(path, encoding="utf8") as f:
            s2r_template_dict = json.load(f)

        og_template: TemplateScan2Report = TemplateScan2Report.parse_dict_custom(fid=fid, locale="og", template_dict=s2r_template_dict, original=True)
        if og_template is None:
            logger.info(f"Empty og template for fid {fid}?")
            continue
        og_template_tns = Scan2ReportConvert.convert_single_finding_to_tns_pwndoc(og_template, "og")

        # todo: fix this - I shouldn't need to hack it here
        og_template_tns.set_is_template(True)
        if og_template_tns.hosts == "":
            og_template_tns.hosts = []

        PwndocTemplateManager.add_single_locale_template(og_template_tns)

    TemplateScan2Report.delete_template_files(["og"])
    PwndocTemplateManager.update_template_db_from_pwndoc()  # Update DB to have a current state.


def _success_msg(ps: ProcessingSettings, prefix: str = "") -> markupsafe.Markup:
    answer: List[str] = []
    if len(prefix):
        answer.append(prefix)

    if ps.upload_to_pwndoc:
        pass  # currently handled in outer scope to have access to more data
    else:
        answer.append(f'<a href="{url_for("api_process_findings.show_file_output", folder_name=ps.get_folder_name())}" download>Result file</a>')

    answer.append(f'<a href="{url_for("api_process_findings.show_meta_output", folder_name=ps.get_folder_name())}">More information</a>')
    # answer.append(f'{len(ps.missing_templates)} findings are missing templates and therefore were not imported')  # todo: change this behaviour

    return markupsafe.Markup("<br>".join(answer))


@bp.route("/process_scanner_result_using_scan2report/<folder_name>", methods=["GET"])
@bp.route("/process_scanner_result_using_scan2report/<folder_name>/", methods=["GET"])
def process_scanner_result_using_scan2report(folder_name: str):
    ps: Optional[ProcessingSettings] = ProcessingSettings.load_from_folder_name(folder_name)
    if ps is None:
        return "Folder doesn't exist", 400

    TemplateScan2Report.delete_template_files([ps.locale, "og"])
    download_templates_from_pwndoc_to_scan2report(
        "scan2report_native" if ps.audit_id == OutputFormat.DOCX.value else "scan2report_with_pwndoc_formatting",
        locale=ps.locale
    )

    files = [os.path.join(ps.get_folder_path(), f) for f in ps.stored_files]
    args = shlex.split(ps.scan2report_args)

    if "--output" in args:
        return "Don't specify --output as a parameter, use selector of audit ID instead.", 400

    output_path = ps.get_result_path()
    combined_args = ['--output', output_path]

    if not ("-l" in args or "--lang" in args):
        combined_args += ['--lang', ps.locale]

    combined_args += args + files

    logger.debug(f"Running scan2report with: {combined_args}")

    # todo: the http connection might time out - do as async?
    count_of_known_templates_before = len(TemplateScan2Report.list_files_in_locale_folder(ps.locale, original=False))

    scan2report.main(combined_args)  # First run to generates OG templates

    # Metadata from the first load - extract the missing templates.
    meta_data = ps.load_meta_file()
    ps.missing_templates = meta_data["missing"]  # _upload_missing_templates
    ps.output_unspecified["missing_n_templates"] = len(ps.missing_templates)
    ps.save()

    _upload_missing_templates(ps)
    download_templates_from_pwndoc_to_scan2report(
        output_format="scan2report_native" if ps.audit_id == OutputFormat.DOCX.value else "scan2report_with_pwndoc_formatting",
        locale='og',
        force_locale_folder=ps.locale,
        prefix_title_with_locale=True
    )

    count_of_known_templates_after = len(TemplateScan2Report.list_files_in_locale_folder(ps.locale, original=False))
    if count_of_known_templates_before != count_of_known_templates_after:
        scan2report.main(combined_args)  # Second run to regenerate the findings if during the first run we found OG templates

    ps.cleanup_uploaded_files()

    if ps.upload_to_pwndoc:
        return redirect(url_for('api_process_findings.upload_scan_findings_to_pwndoc', folder_name=ps.get_folder_name()))

    flash_raw_html = _success_msg(ps)
    return _flash_msg_and_go_to_start(flash_raw_html, "info", ps.audit_id)


def get_changed_after_combine_with_existing_findings(audit_id: str, possibly_new_findings: List[dict]) -> List[dict]:
    existing_raw_findings = get_findings_from_audit(audit_id)

    fid_mapping = {}
    for single_raw_finding in existing_raw_findings:
        fid = get_fid_from_raw_finding(single_raw_finding)
        if fid in fid_mapping:
            FlashLog.warning(f"Multiple existing findings with {fid} in single audit. Overwriting.")
        fid_mapping[fid] = single_raw_finding
        single_raw_finding["_changed"] = False

    for single_raw_finding in possibly_new_findings:
        fid = get_fid_from_raw_finding(single_raw_finding)

        # Interesting tests for the HTML parsing
        # if fid not in ["burp_5243008", "HTTP_generic_XSS"]:
        #     continue

        for attr_name in HTML_ATTRIBUTES:
            single_raw_finding[attr_name] = MdAndHTMLConvertor.html_to_pwndoc_html(single_raw_finding.get(attr_name, ""))

        if fid not in fid_mapping:
            fid_mapping[fid] = single_raw_finding
            single_raw_finding["_changed"] = True
            continue

        original_scope = fid_mapping[fid].get("scope", "").split("<br>")
        new_scope = single_raw_finding.get("scope", "").split("<br>")

        merged_scope = original_scope + new_scope
        merged_scope_deduplicated = list(dict.fromkeys(merged_scope))
        fid_mapping[fid]["scope"] = "<br>".join(merged_scope_deduplicated)

        if len(merged_scope_deduplicated) != len(new_scope):
            fid_mapping[fid]["poc"] = fid_mapping[fid].get("poc", "") + "<br><br>=====<br><br>" + single_raw_finding.get("poc", "")
            single_raw_finding["_changed"] = True

    new_or_changed = list(filter(lambda x: x.get("_changed", True), fid_mapping.values()))
    logger.info(f"Import contained {len(list(fid_mapping.values()))} unique FIDs which resulted in upsert of {len(new_or_changed)} findings.")

    return new_or_changed


def _extend_audit_services(audit_id: str, new_services: dict) -> helpers.table_of_services.Scope:
    answer = helpers.table_of_services.get_tns_table_of_services(audit_id)

    persistent_scopes_file_path = persistent_audit_scopes_file_path(audit_id)
    local_services_copy = json_safe_load(persistent_scopes_file_path)
    if local_services_copy is not None:
        for host in local_services_copy.get('scope', {}).get('hosts', []):
            for service in host.get('services', []):
                answer.add_entry(host['ip'], service['port'], service['protocol'], service['name'])

    for host_ip, port_list in new_services.items():
        for port_number, l3_protocol_list in port_list.items():
            for l3_protocol, app_protocol in l3_protocol_list.items():
                answer.add_entry(host_ip, port_number, l3_protocol, app_protocol)
    return answer


@bp.route("/force_upload_services/<audit_id>", methods=["GET"])
def force_upload_services(audit_id: str):
    services_as_pwndoc_scope = _extend_audit_services(audit_id, {})
    helpers.table_of_services.set_tns_table_of_services(audit_id, services_as_pwndoc_scope)
    return f"OK; {len(services_as_pwndoc_scope.hosts)} hosts are in the table of services"


@bp.route("/upload_scan_findings_to_pwndoc/<folder_name>", methods=["GET"])
def upload_scan_findings_to_pwndoc(folder_name: str):
    ps: Optional[ProcessingSettings] = ProcessingSettings.load_from_folder_name(folder_name)
    if ps is None:
        return "Folder doesn't exist", 400

    path = ps.get_result_path()
    with open(path, encoding="utf8") as f:
        file_data = json.load(f)

    findings: List[TemplateScan2Report] = []
    locale = ps.locale

    for single_finding in file_data:
        finding_parsed = TemplateScan2Report.parse_dict_custom(single_finding["fid"], locale, single_finding)
        if finding_parsed is None:
            logger.warning(f"Empty finding in audit report?")
        else:
            findings.append(finding_parsed)

    login()
    new_raw_findings = []
    for single_finding in findings:
        single_finding_pwndoc = Scan2ReportConvert.convert_single_finding_to_tns_pwndoc(single_finding, locale)
        single_finding_pwndoc.set_custom_field(
            "original_description_from_scan",
            MdAndHTMLConvertor.html_to_pwndoc_html( template_pwndoc.to_html(single_finding.description) )
        )

        single_finding_pwndoc_dict = single_finding_pwndoc.to_pwndoc_dict()
        new_raw_findings.append(single_finding_pwndoc_dict)

    changed_raw_findings = get_changed_after_combine_with_existing_findings(ps.audit_id, new_raw_findings)
    for single_raw_finding in tqdm(changed_raw_findings, desc="Upserting findings"):
        upsert_raw_finding(ps.audit_id, single_raw_finding)

    meta_data = ps.load_meta_file()
    services_as_pwndoc_scope = _extend_audit_services(ps.audit_id, meta_data.get("services", {}))
    helpers.table_of_services.set_tns_table_of_services(ps.audit_id, services_as_pwndoc_scope)

    flash_raw_html = _success_msg(ps, prefix=f'OK, added or updated {len(changed_raw_findings)} findings - you can now find them in PwnDoc.')
    return _flash_msg_and_go_to_start(flash_raw_html, "info", ps.audit_id)


@bp.route("/meta_output/<folder_name>", methods=["GET"])
def show_meta_output(folder_name: str):
    ps: Optional[ProcessingSettings] = ProcessingSettings.load_from_folder_name(folder_name)
    if ps is None:
        return "Folder doesn't exist", 400

    missing_templates_splitted = [x.split("\t") for x in ps.missing_templates]

    return render_template("findings_meta_output.html",
                           ps=ps,
                           missing_templates=missing_templates_splitted,
                           # import_id=folder_name,
                           )


@bp.route("/file_output/<folder_name>", methods=["GET"])
def show_file_output(folder_name: str):
    ps: Optional[ProcessingSettings] = ProcessingSettings.load_from_folder_name(folder_name)
    if ps is None:
        return "Folder doesn't exist", 400
    if ps.get_result_path().endswith(".docx"):  # todo: don't match on extension
        return send_file(ps.get_result_path(), as_attachment=True)

    with open(ps.get_result_path(), "r", encoding="utf8") as f:
        text = f.read()
    text = html.escape(text, quote=False)
    text = MdAndHTMLConvertor.reverse_custom_escaping_of_html(text)

    filename = os.path.basename(ps.get_result_path())

    return send_file(io.BytesIO(text.encode("utf8")), as_attachment=True, attachment_filename=filename)


