from pathlib import Path

from flask import flash
from loguru import logger
import helpers.custom_logging
from flask import Flask, request, render_template, Response
from flask_migrate import Migrate

from template_pwndoc import TNSTemplatePwndoc
from template_manager import PwndocTemplateManager
from helpers.db_models import db, DbTemplate

import config
import pwndoc_db_init
from helpers.disk_file_versioning import DiskFileVersioning
from helpers.file_utils import *
from helpers.template_grouping import TemplateGrouping

from api_process_findings import bp as process_findings_bp
from api_templates import bp as templates_bp
from api_debug import bp as debug_bp

logger.add("user_data/importer-logs/main.log", rotation="1 week", retention="60 days", compression="gz", encoding="utf8")

pwndoc_db_init.InitialData()

app = Flask(__name__)

app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///user_data/importer-db/main.db'
app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
db.init_app(app)

migrate = Migrate()

# https://github.com/miguelgrinberg/Flask-Migrate/issues/61#issuecomment-208131722
with app.app_context():
    if db.engine.url.drivername == 'sqlite':
        migrate.init_app(app, db, render_as_batch=True, compare_type=True)
    else:
        migrate.init_app(app, db)

with app.app_context():
    db.create_all()

app.config['UPLOAD_FOLDER'] = relative_path('debug_tmp')
app.config['MAX_CONTENT_LENGTH'] = 100 * 1000 * 1000
app.config["JSONIFY_PRETTYPRINT_REGULAR"] = True
app.config["SECRET_KEY"] = config.SECRET_KEY


Path(app.config['UPLOAD_FOLDER']).mkdir(parents=True, exist_ok=True)

app.register_blueprint(process_findings_bp, url_prefix='/findings')
app.register_blueprint(templates_bp, url_prefix='/templates')

# Todo: possibly disable this or require additional password
app.register_blueprint(debug_bp, url_prefix='/debug')

TemplateGrouping.add_new_gids()


@app.route("/")
def homepage():
    return render_template('base.html')


@app.errorhandler(500)
def internal_error(error):
    flash(f"Internal server error (more information is in server logs)", "danger")
    return render_template('base.html'), 500


@app.errorhandler(404)
def internal_error(error):
    flash(f"The requested URL doesn't exist.", "danger")
    return render_template('base.html'), 404


@app.route("/json_editor", methods=["GET"])
def json_editor():
    PwndocTemplateManager.update_template_db_from_pwndoc()
    files_selector = [(x, x) for x in DiskFileVersioning().list_available_files()]
    templates: List[TNSTemplatePwndoc] = PwndocTemplateManager.get_all_templates()
    templates.sort(key=lambda x: x.fid)
    return render_template("json_editor.html", files=files_selector, templates=templates)


@app.route("/versioned_file/<filename>", methods=["GET"])
def get_versioned_file(filename: str):
    file_content = DiskFileVersioning().get_file_from_disk(filename)
    if file_content is None:
        return "Filename doesn't match existing mapping", 404
    return Response(file_content, mimetype='application/json')  # todo: this might not have to be JSON


@app.route("/versioned_file/<filename>", methods=["POST"])
def upload_versioned_file(filename: str):
    data = request.data

    file_content = DiskFileVersioning().save_file_to_disk(filename, data.decode())
    if file_content is None:
        return "Filename doesn't match existing mapping", 404

    if filename == 'groups':
        TemplateGrouping.add_new_gids()
    return "OK"


