import time

import requests
import urllib3
from tqdm import tqdm
from loguru import logger
from flask import abort
from typing import List, Tuple

from helpers.db_models import DbPwndocMapping
from template_pwndoc import TemplatePwndoc, MultilocaleTemplateWrapperPwndoc, TNSTemplatePwndoc, \
    to_html

from config import PWNDOC_URL, PWNDOC_USERNAME, PWNDOC_PASSWORD, PWNDOC_DISABLE_HTTPS_VERIFICATION
from helpers.file_utils import *

session = requests.Session()
session.verify = not PWNDOC_DISABLE_HTTPS_VERIFICATION
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)  # I'm writing one custom warning on config load, the default setting of urllib3 prints it on every request.

NO_FID = "NO_FID"


def login():
    data = {
        "username": PWNDOC_USERNAME,
        "password": PWNDOC_PASSWORD,
        "totpToken": "",
    }

    resp = session.post(f"{PWNDOC_URL}/api/users/token", data=data)
    assert resp.status_code == 200, f"Login failed. Did you set PwnDoc username ({PWNDOC_USERNAME}) and password ({'*'*len(PWNDOC_PASSWORD)}) in ENV variables?"


def refresh_token():
    # todo: maybe use requests-oauthlib instead?
    # https://requests-oauthlib.readthedocs.io/en/latest/examples/real_world_example_with_refresh.html
    resp = session.get(f"{PWNDOC_URL}/api/users/refreshtoken")
    assert resp.status_code == 200


def add_finding_to_audit(finding: 'TemplatePwndoc', audit_id: str):
    data = finding.to_pwndoc_dict()
    upsert_raw_finding(audit_id, data)


def upsert_raw_finding(audit_id: str, data: dict):
    finding_id = data.get("_id")
    if finding_id is None:
        url = f"{PWNDOC_URL}/api/audits/{audit_id}/findings"
        resp = session.post(url, json=data)
    else:
        url = f"{PWNDOC_URL}/api/audits/{audit_id}/findings/{finding_id}"
        resp = session.put(url, json=data)
    assert resp.status_code == 200, f"Upserting of finding failed; {resp.text}"


def add_templates_for_vulnerability(lang_mutations_of_template: Dict[str, 'TemplatePwndoc']):
    """ Don't use this directly, rather use PwndocTemplateManager.add_single_locale_template """

    url = f"{PWNDOC_URL}/api/vulnerabilities"

    data = MultilocaleTemplateWrapperPwndoc()

    if len(lang_mutations_of_template) == 0:
        logger.warning(f"Skipping empty dictionary of templates.")
        return

    fid = NO_FID
    for locale, single_template in lang_mutations_of_template.items():
        fid = getattr(single_template, "fid", fid)
        single_template.set_is_template(True)

        og_template: Optional[TNSTemplatePwndoc] = lang_mutations_of_template.get("og")
        if og_template and len(single_template.description):
            single_template.set_custom_field("original_description_from_template", to_html(single_template.description))

        data.details.append(single_template)

    if fid != NO_FID:
        pwndoc_template_fid = DbPwndocMapping.get_pwndoc_id_from_fid(fid)
        if pwndoc_template_fid:
            data._id = pwndoc_template_fid

    pwndoc_json_format = data.get_pwndoc_repr()

    if data._id is None:
        resp = session.post(url, json=[pwndoc_json_format])
    else:
        resp = session.put(f"{url}/{data._id}", json=pwndoc_json_format)

    if resp.status_code == 422 and resp.json()["datas"] == "Vulnerability title already exists":
        logger.warning(f"Rewrite of existing vulnerability '{fid}' is not supported. Skipping.")
    elif resp.status_code in [200, 201]:
        # todo: saving the pwndoc ID mapping here would be nice, though it's not inside the resp
        pass
    else:
        assert False, f"Creation of vulnerability failed: {resp.status_code} {resp.text}"


def filename_from_url_path(url_path: str):
    return f"{url_path.replace('/', '_')}.json"


def get_and_save(url_path: str, filename: Optional[str] = None) -> dict:
    if filename is None:
        filename = filename_from_url_path(url_path)

    url = f"{PWNDOC_URL}{url_path}"
    resp = session.get(url)
    if resp.status_code != 200:
        FlashLog.error(f"PwnDoc API failed {url_path} | {resp.status_code} | {resp.text}")
        abort(500)

    def _save_to_file(data: dict, filename: str):
        with open(relative_path(f"api_examples/{filename}"), "w", encoding="utf8") as f:
            json.dump(data, f, indent=4)

    _save_to_file(resp.json(), filename)
    if url_path == "/api/vulnerabilities":
        _save_to_file(resp.json(), filename+f"_backup_{time.time()}.json")
    return resp.json()


def pwndoc_download_json_of_vulnerabilities_and_custom_fields():
    """ Don't use this directly, use PwndocTemplateManager.update_template_db_from_pwndoc instead. """
    login()
    get_and_save("/api/vulnerabilities")
    get_and_save("/api/data/custom-fields")


def pwndoc_download_json_of_custom_fields():
    login()
    get_and_save("/api/data/custom-fields")


def _download_audits():
    login()
    get_and_save("/api/audits")


def refresh_examples(audit_id: Optional[str] = None):
    urls = [
        "/api/vulnerabilities",
        "/api/settings/public",
        "/api/audits",
        "/api/data/languages",
        "/api/data/audit-types",
        "/api/data/custom-fields",
        "/api/data/vulnerability-categories",
        "/api/data/sections",
        "/api/companies",
        "/api/clients",
        "/api/templates",
        "/api/users",
        "/api/users/reviewers",
    ]
    print("Updating API examples")
    for url in tqdm(urls):
        get_and_save(url)

    if audit_id:
        download_audit(audit_id)


def load_audit_file(audit_id: str) -> Optional[dict]:
    """ Note: Use refresh_examples(audit_id) before hand. """
    filepath = relative_path(f"api_examples/{filename_from_url_path('/api/audits/'+audit_id)}")
    data = json_safe_load(filepath)
    return data


def delete_all_templates_in_pwndoc():
    login()
    url = f"{PWNDOC_URL}/api/vulnerabilities"
    resp = session.delete(url)
    assert resp.status_code == 200, f"Template removal failed: {resp.text}"


def download_audit(audit_id: str):
    urls = [
        f"/api/audits/{audit_id}",
        f"/api/audits/{audit_id}/general",
        f"/api/audits/{audit_id}/network",
    ]
    for url in urls:
        get_and_save(url)


def download_and_return_audit(audit_id: str) -> dict:
    login()
    download_audit(audit_id)
    answer = load_audit_file(audit_id)
    if answer is None:
        abort(400, "Audit with this ID doesn't exist.")
    return answer.get('datas', {})


def get_findings_from_audit(audit_id: str) -> List[dict]:
    old_audit = download_and_return_audit(audit_id)

    # note: Findings can be extracted from the Audit endpoint, however if I take the finding from the Audit endpoint
    #   and upsert it using the Finding endpoint, then it breaks custom fields in Pwndoc UI.
    #   Getting the data from Finding endpoint is less efficient, but works.
    findings_from_audit_format = old_audit.get("findings", [])
    findings_from_finding_format = []

    for single_finding in findings_from_audit_format:
        data = download_and_return_finding(audit_id, single_finding.get("_id"))
        findings_from_finding_format.append(data)
    return findings_from_finding_format


def download_and_return_finding(audit_id: str, finding_id: str):
    # login()
    url_path = f"/api/audits/{audit_id}/findings/{finding_id}"
    return get_and_save(url_path).get("datas", {})


def delete_finding(audit_id: str, finding_id: str):
    login()
    url_path = f"/api/audits/{audit_id}/findings/{finding_id}"
    resp = session.delete(f"{PWNDOC_URL}{url_path}")
    assert resp.status_code == 200, f"Template removal failed: {resp.text}"


def update_scope(audit_id: str, scope_raw: dict):
    login()
    url_path = f"/api/audits/{audit_id}/network"
    resp = session.put(f"{PWNDOC_URL}{url_path}", json=scope_raw)

    json_dump(scope_raw, persistent_audit_scopes_file_path(audit_id))

    assert resp.status_code == 200, f"Scope update failed: {resp.text}"


def _get_audits_data() -> List[dict]:
    _download_audits()
    with open("api_examples/_api_audits.json", encoding="utf8") as f:
        audits = json.load(f)
    return audits["datas"]


def get_audits() -> List[Tuple[str, str]]:
    answer = []
    for x in _get_audits_data():
        answer.append((x["_id"], x["name"]))
    return answer


# pwndoc_download_vulnerabilities_and_custom_field()
