import base64
import random

import requests
from tqdm import tqdm
from loguru import logger
import time
import json

from config import PWNDOC_URL, PWNDOC_USERNAME, PWNDOC_PASSWORD
from helpers.file_utils import relative_path

from pwndoc_api import session, login


class InitialData:
    __NEW_MAPPING_FILEPATH = relative_path(f'user_data/pwndoc-init/new_id_mapping.json')
    __INIT_INDICATOR_NAME = "INIT_STATUS"
    __INIT_FINISHED = "INIT_FINISHED"

    def __init__(self, force=False):
        time.sleep(random.uniform(0, 1))  # randomly offset threads to not overload pwndoc

        self._save_new_id_mapping("NO_CONTENT", "LOREM_IPSUM")  # force file creation if doesn't exist
        was_db_already_initialized = self._load_new_id_mapping(self.__INIT_INDICATOR_NAME) == self.__INIT_FINISHED

        is_db_clean = self._is_db_clean()
        force |= is_db_clean

        if was_db_already_initialized and not force:
            logger.info("DB seems to be (at least partially) initialized - skipping initialization.")
            return

        self._add_pwndoc_user(first_user=True)
        login()

        # no dependencies
        self._upload_universal('_api_templates.json')
        self._upload_universal('_api_data_languages.json')
        self._upload_universal('_api_data_sections.json')
        self._upload_universal('_api_data_vulnerability-categories.json')
        self._upload_universal('_api_data_audit-types.json')  # requires: templates, languages, sections
        self._upload_universal('_api_data_custom-fields.json')  # requires: everything

        # api_templates.import_scan2report_templates_to_pwndoc()

        self._save_new_id_mapping(self.__INIT_INDICATOR_NAME, self.__INIT_FINISHED)
        logger.info("DB Init completed")

    def _is_db_clean(self):
        for i in range(10):
            try:
                resp = session.get(f"{PWNDOC_URL}/api/users/init")
                if resp.status_code == 200 and resp.json().get("datas") == True:
                    logger.info("Clean instance detected. Forcing import to DB.")
                    return True
                return False
            except requests.exceptions.ConnectionError:
                logger.warning(f"Connection to {PWNDOC_URL} failed. Waiting before retry.")
                time.sleep(random.uniform(1, 2))

        return False

    @staticmethod
    def _add_pwndoc_user(username: str = PWNDOC_USERNAME, password: str = PWNDOC_PASSWORD, firstname: str = "Pwndoc",
                         lastname: str = "Importer", first_user: bool = False):

        init_data = {
            "username": username,
            "password": password,
            "firstname": firstname,
            "lastname": lastname
        }
        url = f"{PWNDOC_URL}/api/users"

        if first_user:
            url += "/init"
        else:
            init_data["role"] = "admin"

        resp = session.post(url, data=init_data)
        if resp.status_code == 403 and resp.json()["datas"] == "Already Initialized":
            logger.info("Init failed - already initialized, possibly by another thread. Skipping.")
            return
        elif resp.status_code == 422 and resp.json()["datas"] == "Username already exists":
            logger.info(f"User '{username}' already exists. Skipping.")
            return
        assert resp.status_code == 201, f"Init failed: {resp.text}"

    def _save_new_id_mapping(self, old_id: str, new_id: str):
        try:
            with open(self.__NEW_MAPPING_FILEPATH, encoding="utf8") as f:
                content = json.load(f)
        except:
            content = {}

        content[old_id] = new_id
        with open(self.__NEW_MAPPING_FILEPATH, "w", encoding="utf8") as f:
            json.dump(content, f, indent=4)

    def _load_new_id_mapping(self, old_id: str) -> str:
        for i in range(10):
            try:
                with open(self.__NEW_MAPPING_FILEPATH, encoding="utf8") as f:
                    content = json.load(f)
                return content.get(old_id, old_id)
            except json.decoder.JSONDecodeError:
                logger.info("Loading for new ID mapping failed. Sleeping before retry.")
                time.sleep(random.uniform(0, 1))

        assert False, "Loading for new ID mapping failed - even on final retry. Killing."

    @staticmethod
    def _get_api_path_from_filename(filename: str) -> str:
        url_path = filename.replace("_", "/").rsplit(".", 1)[0]
        return f"{PWNDOC_URL}{url_path}"

    @staticmethod
    def _get_data_from_init_file(filename: str) -> dict:
        with open(relative_path(f'user_data/pwndoc-init/{filename}'), encoding="utf8") as f:
            content = json.load(f)["datas"]
        return content

    def _upload_universal(self, filename: str):
        content = self._get_data_from_init_file(filename)

        for single_item in content:

            if filename == "_api_templates.json":
                single_item['file'] = self._load_template_file(single_item['name'] + "." + single_item['ext']).decode()

            if filename == "_api_data_audit-types.json":
                for locale_dict in single_item.get("templates", []):
                    locale_dict["template"] = self._load_new_id_mapping(locale_dict["template"])

            resp = session.post(self._get_api_path_from_filename(filename), json=single_item)
            if resp.status_code == 201:
                if single_item.get("_id"):
                    self._save_new_id_mapping(single_item["_id"], resp.json()["datas"]["_id"])
            else:
                if not resp.json().get("data", "").endswith("already exists"):
                    logger.info(f"Skipping upload duplicate: {resp.text}")
                else:
                    logger.warning(f"Upload failed: {resp.text}. Skipping.")

    @staticmethod
    def _load_template_file(filename: str) -> bytes:
        template_filepath = f"user_data/pwndoc-init/{filename}"
        with open(relative_path(template_filepath), "rb") as f:
            template_file = f.read()
            template_file_b64 = base64.b64encode(template_file)
            return template_file_b64
