from collections import defaultdict
import json
from typing import List, Optional, Dict

import pwndoc_api
from helpers.db_models import DbTemplate, DbPwndocMapping
from pwndoc_api import pwndoc_download_json_of_vulnerabilities_and_custom_fields
from pwndoc_custom_fields import PwndocCustomFields
from template_pwndoc import MultilocaleTemplateWrapperPwndoc, TNSTemplatePwndoc


class PwndocTemplateManager:

    @classmethod
    def update_template_db_from_pwndoc(cls):
        pwndoc_download_json_of_vulnerabilities_and_custom_fields()
        current_multilocale_templates_in_pwndoc: List[MultilocaleTemplateWrapperPwndoc] = cls._parse_pwndoc_multilang_templates()

        for single_fid_templates in current_multilocale_templates_in_pwndoc:
            for single_template in single_fid_templates.details:
                single_template: TNSTemplatePwndoc = single_template
                DbTemplate.add_template(single_template.fid, single_template.locale, single_template)

    @staticmethod
    def _parse_pwndoc_multilang_templates() -> List[MultilocaleTemplateWrapperPwndoc]:
        with open("api_examples/_api_vulnerabilities.json", encoding="utf8") as f:
            data = json.load(f)

        pwndoc_templates: List[dict] = data["datas"]

        answer = []
        for x in pwndoc_templates:
            answer.append(PwndocTemplateManager.__parse_single_pwndoc_template(x))

        return answer

    @staticmethod
    def get_all_templates() -> List[TNSTemplatePwndoc]:
        return [x.decode_content() for x in DbTemplate.get_all_current_templates()]

    @classmethod
    def get_all_templates_as_dict_of_locales(cls) -> Dict[str, Dict[str, TNSTemplatePwndoc]]:
        all_templates = cls.get_all_templates()
        dict_of_fids_and_locales: Dict[str, Dict[str, TNSTemplatePwndoc]] = defaultdict(lambda: defaultdict(TNSTemplatePwndoc))
        for single_template in all_templates:
            dict_of_fids_and_locales[single_template.fid][single_template.locale] = single_template
        return dict_of_fids_and_locales

    @staticmethod
    def find_pwndoc_id_by_fid(fid: str) -> Optional[str]:
        return DbPwndocMapping.get_pwndoc_id_from_fid(fid)

    @staticmethod
    def __parse_single_pwndoc_template(template: dict) -> MultilocaleTemplateWrapperPwndoc:  # returns all lang mutations
        pwndoc_id = template.get("_id", None)
        multilang_template = MultilocaleTemplateWrapperPwndoc(_id=pwndoc_id)

        for lang_dict in template["details"]:
            tmp = PwndocTemplateManager.__parse_single_pwndoc_locale(lang_dict)
            multilang_template.details.append(tmp)

        if multilang_template.get_fid():
            DbPwndocMapping.save_fid_and_pwndoc_id(multilang_template.get_fid(), pwndoc_id)

        return multilang_template

    @staticmethod
    def __parse_single_pwndoc_locale(lang_dict: dict) -> TNSTemplatePwndoc:  # returns all lang mutations
        answer = {}

        for field in lang_dict:
            if field == "customFields":
                continue
            answer[field] = lang_dict[field]

        for custom_field in lang_dict["customFields"]:
            custom_field_id = custom_field["customField"]
            custom_field_label = PwndocCustomFields().get_custom_field_by_id(custom_field_id)["label"]

            answer[custom_field_label] = custom_field["text"]
            if isinstance(answer[custom_field_label], dict):
                assert len(answer[custom_field_label]) == 2, "Dict doesn't contain value and locale?"
                answer[custom_field_label] = answer[custom_field_label]['value']

        tmp = TNSTemplatePwndoc.from_dict(answer)
        tmp.set_is_template(True)

        return tmp

    @classmethod
    def add_locales_for_fid(cls, templates: List[TNSTemplatePwndoc]) -> bool:
        some_new = False
        fid = templates[0].fid

        for template in templates:
            _, is_new = DbTemplate.add_template(template.fid, template.locale, template)
            some_new |= is_new

        if some_new:
            dict_of_current_templates = cls.get_all_templates_as_dict_of_locales()
            pwndoc_api.add_templates_for_vulnerability(dict_of_current_templates[fid])

        return some_new

    @classmethod
    def add_single_locale_template(cls, template: TNSTemplatePwndoc) -> bool:
        return cls.add_locales_for_fid([template])
